/* N-body calculation using Greenspan 1st-order discrete mechanics.    */
/* Incorporates xyz-parallelism, uses domain pointers, combines        */
/* particle and pair domains, diagonal processors store particles      */
/* ***  Includes communication operations only - arithmetic omitted    */
/* Spreading coded in C*, reduced overhead in accumulation,            */
/*  more globals, Booleans stored in char fields.                      */

#include <stdio.hs>
/* #include <math.hs> */
#include "cmtime.hs"

/* Max number of bodies in the calculation.  N = 64 fits in 16k procs
   N must be a power of 2, 4N^2 <= no. of v. procs. */
#define N 64

int NB;      /* Number of bodies in system, <= N */
int LNB;     /* Largest power of 2 < NB */

/* Particles, and pairs of particles.  Diagonal entries have particle	*/
/* states.  Some variables could be overloaded. 			*/
/* act  - processor is active pair					*/
/* sm   - processor participates in acceleration sum			*/
/* here - processor represents particle					*/
/* sp   - processor participates in spreading                           */
/* smk  - control for spreading, accumulation, when procs change        */
/* th   - 'this', for convenience                                       */

domain pair { int i, j, smk, th;  char act, sm, here, sp;
	      float x0, x1, v0, v1, m, d0, r0, ac;
	      domain pair *pi, *pj;  domain pair *r; };

domain pair P[3][N][N];

/* Pairs will be represented symmetrically.  The function setup
   (below) initializes these so that pi, pj point to appropriate
   particle in corresponding coordinate.   */

void pair::setup()
 { int offset, c;  char smm;  mono int k;
   for ( LNB = 1; LNB < NB; LNB <<= 1 ) ;
   sp = (i<LNB) && (j < LNB);  LNB >>= 1;
   th = (int) this;
   offset = (this - &P[0][0][0]);
   c = offset/(N*N);  i = (offset/N)%N;  j = offset%N;
   sm = (i < NB) && (j < NB);
   act = sm && (i != j);  here = sm && (i == j);
   pi = &P[c][i][i];  pj = &P[c][j][j];
   r = (domain pair *)( (c<2) ? th+N*N : th-2*N*N );
   smm = 1;  smk = i^j;  /* xor */
   for ( k = 1; k < NB; k <<= 1 )
    if (smm && (k & smk)) { smk = k;  smm = 0; }
 }

void prsmk()
 { int u, v;
   [domain pair].
   for ( u = 0; u < 2*LNB; u++)
    { for ( v = 0; v < 2*LNB; v++) printf("%2d ", P[0][u][v].smk);
      printf("\n"); }
 }

void pair::prpflt(x)
 float x;
 { mono int u, v;
   for ( u = 0; u < 2*LNB; u++)
    { for ( v = 0; v < 2*LNB; v++) printf("%9.2e ", P[0][u][v].x);
      printf("\n"); }
 }

/* Load inital positions, velocities, also masses, gravitational const. */

void init()
 { [domain pair].
    if (here) 
     { x0 = 17.5;  x1 = -23.37;  v0 = 5.14;  m = 25.6;  d0 = 14.0;
       r0 = 3.6;  ac = 4.76; }
 }

/* smk[i,j] is a power of 2 corresponding to the stage at which proc[i,j]
   sends it's acceleration to partner and drops out of accumulation. It
   also corresponds to the stage at which proc[i,j] begins to send values
   during spreading operations.                                          */

/* Spread xv from diagonal along i-axis.         */
/* May spread to power of 2 just larger than NB  */

void pair::spreadi(xv)
  float xv;
  { char smm;  mono int k, b;
    smm = here;  
    for ( k = 2*LNB, b = LNB*N;  k > 1;  k >>= 1, b >>= 1 )
     if ( smm || ( k == smk) )
      { smm = 1;  ((domain pair *)(th^b))->xv = xv; }
  }

/* Spread xv from diagonal along j-axis.         */
/* May spread to power of 2 just larger than NB  */

void pair::spreadj(xv)
  float xv;
  { char smm;  mono int k;
    smm = here;  
    for ( k = 2*LNB; k > 1; )
     if ( smm || (k == smk) )
      { smm = 1;  k >>= 1;  ((domain pair *)(th^k))->xv = xv; }
  }

/* Calculate accelerations - called in sp's active */

void pair::accel()
 {    float d1, r1;  char smm;
      mono int k;
      ac = 0.0;  d1 = x1;  spreadi(d1);  spreadj(x1);
      prpflt(d1);  prpflt(x1);
      if (act)
      { /* d1 = d1 - x1; */
        r1 = d1;  ac = r->r1;  r1 = r->ac; };  /* end if (act) ... */
      /* Sum acceleration terms in time ~ O(log N) */
      smm = 1;
      for ( k = 1; k < NB; k <<= 1 )
	{ if ( smm && (k == smk) )
	     { ((domain pair *)(th^k))->ac = ac;  smm = 0; }
	  /* else add */ }   /* end for */
 }

/* Calculate initial accelerations */

void pair::accel1()
 {    char smm;  mono int k;
      ac = 0.0;  d0 = x0;  spreadi(d0);  spreadj(x0);
      prpflt(d0);  prpflt(x0);
      if (act)
      { /* d0 = d0 - x0; */
	r0 = d0;  ac = r->r0;  r0 = r->ac; };  /* end if (act) */
      /* Sum acceleration terms in time ~ O(log N) */
      smm = 1;
      for ( k = 1; k < NB; k <<= 1 )
	{ if ( smm && (k == smk) )
	     { ((domain pair *)(th^k))->ac = ac;  smm = 0; }
	  /* else add */ }   /* end for */
 }


/* Greenspan's 1st-order method.  Performs ns steps of time dt
   with solution tolerance eps.  Returns number of evaluations.
   Leaves number of iterations which were stopped at limit in cfail.
   Some printing for debugging is provided in comments.  */

int cfail;   /* Counts number of iterations that failed to converge */

int green1(ns, lim, dt, eps)
  int ns, lim;  float dt, eps;
 { int k, ic, v;  float hdt, mx, my;
   v = 0;  hdt = 0.005;
   [domain pair]. if (sp)  /* Everything in "spreading" procs */
   { float nx;
     if (act) m = pj->m;
     for (k = 0; k < ns; k++)
       { if (here) { x1 = x0;  v1 = v0;  ic = 0; };
         accel1();
	 /* This is a do - while - repeat loop: */
   	 do { v++;  ic++;
	      if ( ! ( ic < lim ) ) break;
	      accel(); }
	 while (1);
       } ;   /* end for */
    }  /* end  [domain pair] */
   return(v);
 }

/* Commands are:
   r - run (prompts for parameters)
   p - print smk
   q - quit                         */

void main()
 { int ns, np, nl, nv, k, vl;  float dt, eps;  long tm;
   char c[2];
   while ( printf("nbcmtim> "), scanf("%1s", c), c[0] != 'q' )
    if ( c[0] == 'p' ) prsmk(); else
    {
     /* Performs np integrations, each ns steps of dt */
     printf(
         "Enter NB, no. of evals per step, no. of steps, no. of cycles.\n> ");
     scanf("%d%d%d%d", &NB, &nv, &ns, &np);
     [domain pair].{ setup(); };
     init();
     printf("\n time = 0");
     for ( k = 0; k < np; k++ )
      { CM_start_timer(1);
        cfail = 0;
        vl = green1(ns, nv, dt, eps);
        CM_stop_timer(1);  CM_reset_timer();
        printf(
	 "\n simulated time = %9.3f  %d evaluations  %d iterations limited\n",
	  (k + 1)*ns*dt, vl, cfail);
      };
    }
 }

